using System;
using System.Collections;
using System.Drawing;
using Microsoft.Win32;

using SdlDotNet;
using Tao.OpenGl;

namespace Jumpers
{
	/// <summary>
	/// Game contstants.
	/// </summary>
	public sealed class GameConstants
	{
		public static readonly float LightRadius = 200.0f;

		public static readonly float TrampolineWidth = 200.0f;

		public static readonly float JumperRadius = 5.0f;

		private GameConstants()
		{
		}
	}

	/// <summary>
	/// The players trampoline
	/// </summary>
	public class Trampoline
	{
		public Vector2 pos;
		public int pos_idx;
		public float h;
		public float hvel;

		private float[] m_positions = new float[ 3 ];

		public Trampoline()
		{
			pos = new Vector2( -320, 460 );
			pos_idx = 1;
			h = 10;
			hvel = 0;

			// Calculate the positions the trampoline can be in
			float w = GameConstants.TrampolineWidth / 2.0f;
			m_positions[ 0 ] = 85.0f + w;
			m_positions[ 2 ] = 635.0f - w;
			m_positions[ 1 ] = ( m_positions[ 0 ] + m_positions[ 2 ] ) / 2.0f;

			pos.X = m_positions[ pos_idx ];
		}

		public void Update()
		{
			float w = GameConstants.TrampolineWidth / 2.0f;

			float new_pos = m_positions[ pos_idx ];
			pos.X = 0.7f * pos.X + 0.3f * new_pos;

			// Bounciness
			h += hvel;

			if ( h < 0 ) { h = 0; hvel = 0; }
			if ( h > 20 ) { h = 20; hvel = 0; }

			float hdir = 10 - h;
			if ( Math.Abs( hdir ) < 0.1f )
			{
				h += hdir;
				hdir = 0.0f;
			}
//			hdir = ( hdir < 0 ) ? -1.0f : 1.0f;

			hvel += hdir * 0.01f;
			hvel *= 0.98f;
		}

		public bool Bounce( Vector2 jpos )
		{
			float y = 460 - h - GameConstants.JumperRadius;

			if ( jpos.Y < y )
				return false;

			// Inside in X-axis?
			float w = 3.0f + GameConstants.TrampolineWidth / 2.0f;
			Vector2 diff = jpos - pos;
			if ( diff.X < -w || diff.X > w )
				return false;

			// Bounce
			hvel -= 0.5f;
			return true;
		}

		public void Draw()
		{
			// Draw it
			Gl.glPushMatrix();
			Gl.glTranslatef( pos.X, pos.Y, 0 );

			Gl.glLineWidth( 2.0f );

			Gl.glBegin( Gl.GL_LINES );
			Gl.glColor3f( 0.8f, 0.83f, 0.85f );

			// (spring)
			Gl.glVertex2f( 0, 0 );
			Gl.glVertex2f( 0, -h );

			// (body)
			float w = GameConstants.TrampolineWidth / 2.0f;
			Gl.glVertex2f( -w, -h );
			Gl.glVertex2f( w, -h );

			Gl.glEnd();

			Gl.glBlendFunc( Gl.GL_ONE, Gl.GL_ZERO );

			Gl.glPopMatrix();
		}
	}

	/// <summary>
	/// The main game screen.
	/// </summary>
	public class GameScreen : IScreen
	{
		private Vector2 m_light_pos;

		private Trampoline m_trampoline = new Trampoline();

		private ParticleSystem m_particles = new ParticleSystem();

		private ArrayList m_jumpers = new ArrayList();
		private int m_jumper_pause;
		private int m_jumper_pause_max;

		private float m_extra_radius;
		private float m_extra_radius_dest;

		private int m_game_over_countdown = 60;
		private bool m_game_over_mouse_down = false;

		private int m_extras_given = 0;

		private int m_score = 0;
		private int m_high = 0;
		private int m_lives = 5;
		private int m_bounces = 0;
		private int m_saved = 0;

		private RegistryKey reg_key;

		public GameScreen()
		{
			m_jumper_pause = 60;
			m_jumper_pause_max = (int) (60 * 4.5f);

			// Get the highscore
			reg_key = Registry.CurrentUser.OpenSubKey( "Software\\LD48H6_Jumpers", true );
			if ( reg_key == null ) 
				reg_key = Registry.CurrentUser.CreateSubKey( "Software\\LD48H6_Jumpers" );

			try { m_high = int.Parse( "" + reg_key.GetValue( "Highscore" ) ); }
			catch { m_high = 0; }
		}

		public void Update(SdlDotNet.Mouse mouse)
		{
			m_light_pos = new Vector2( mouse.MousePosition );

			m_extra_radius = (0.7f) * m_extra_radius + (0.3f) * m_extra_radius_dest;
			m_extra_radius_dest *= 0.8f;

			// Game over?
			if ( m_lives <= 0 &&
				 (m_game_over_countdown--) <= 0 )
			{
				bool is_down = mouse.IsButtonPressed( MouseButton.PrimaryButton );

				if ( !is_down && m_game_over_mouse_down )
				{
					MainLoop.Instance.GameScreen = new MenuScreen();
				}

				m_game_over_mouse_down = is_down;
			}

			// Spawn a new jumper?
			if ( m_lives > 0 &&
				 m_jumper_pause-- < 0 )
			{
				m_jumper_pause = m_jumper_pause_max;
				m_jumpers.Add( new Jumper() );

				m_extra_radius_dest = 20.0f;

				// Decrease the max pause
				int dec = m_jumper_pause_max / 45;
				if ( dec <= 0 )
					dec = 1;
				m_jumper_pause_max -= dec;

				if ( m_jumper_pause_max < 30 )
					m_jumper_pause_max = 30;
			}

			// Update the trampoline
			m_trampoline.Update();

			// Update the jumpers
			int idx = 0;
			while ( idx < m_jumpers.Count )
			{
				Jumper jumper = (Jumper) m_jumpers[ idx ];
				JumperStatus status = jumper.Update( m_trampoline );

				// Update lives/score
				if ( m_lives > 0 )
				{
					switch ( status )
					{
						case JumperStatus.Killed:
							--m_lives;
							break;

						case JumperStatus.Bounced:
							m_score += 1;
							m_bounces += 1;
							break;

						case JumperStatus.OffScreen:
							m_score += 10;
							m_saved += 1;
							break;
					}
				}

				// Spawn particles?
				if ( status == JumperStatus.Killed )
				{
					int pamt = Resources.random.Next( 30, 150 );
					Vector2 ppos = new Vector2( jumper.pos.X, 460 );

					float speed = Math.Abs( jumper.vel.Y ) / 5.0f;
					if ( speed < 0.3f )
						speed = 0.3f;

					for ( int i = 0; i < pamt; ++i )
					{
						m_particles.SpawnParticle( ppos, speed );
					}
				}

				// Remove/continue
				bool remove = (status == JumperStatus.Killed) || (status == JumperStatus.OffScreen);
				if ( remove )
					m_jumpers.RemoveAt( idx );
				else
					++idx;
			}

			// Update the particles
			m_particles.Update();

			// Give an extra life?
			int extra_target = m_score / 100;
			if ( extra_target > m_extras_given )
			{
				m_lives += extra_target - m_extras_given;
				m_extras_given = extra_target;
			}

			// New highscore?
			if ( m_score > m_high )
			{
				m_high = m_score;

				try { reg_key.SetValue( "Highscore", "" + m_high ); }
				catch { }
			}
		}

		public void Draw()
		{
			// Draw the sky
			Gl.glBegin( Gl.GL_TRIANGLE_STRIP );
			Gl.glColor3f( 0.0f, 0.2f, 0.3f );
			Gl.glVertex2f( 0, 0 );

			Gl.glColor3f( 0.0f, 0.05f, 0.1f );
			Gl.glVertex2f( 0, 480 );
			Gl.glColor3f( 0.15f, 0.0f, 0.25f );
			Gl.glVertex2f( 640, 0 );
			Gl.glColor3f( 0.0f, 0.2f, 0.2f );
			Gl.glVertex2f( 640, 480 );
			Gl.glEnd();

			// Draw the light
			Gl.glPushMatrix();
			Gl.glTranslatef( m_light_pos.X, m_light_pos.Y, 0 );

			Gl.glBlendFunc( Gl.GL_ONE, Gl.GL_ONE );

			Gl.glBegin( Gl.GL_TRIANGLE_FAN );

			float pulse_amt = Math.Abs( m_extra_radius / 15.0f );
			if ( pulse_amt > 1.0f )
				pulse_amt = 1.0f;

			float[] color_normal	= new float[] { 0.8f, 0.8f, 0.8f };
			float[] color_pulse		= new float[] { 1.0f, 0.6f, 0.6f };
			Gl.glColor3f(
				pulse_amt * color_pulse[ 0 ] + (1-pulse_amt) * color_normal[ 0 ],
				pulse_amt * color_pulse[ 1 ] + (1-pulse_amt) * color_normal[ 1 ],
				pulse_amt * color_pulse[ 2 ] + (1-pulse_amt) * color_normal[ 2 ] );
			Gl.glVertex2f( 0, 0 );

			Gl.glColor3f( 0.0f, 0.0f, 0.0f );

			for ( int i = 0; i <= 64; ++i )
			{
				float a = 2.0f * (float)Math.PI * ( (float) i / 64.0f );
				float r = GameConstants.LightRadius + m_extra_radius;

				Gl.glVertex2d( Math.Cos( a ) * r, Math.Sin( a ) * r );
			}
			Gl.glEnd();

			Gl.glBlendFunc( Gl.GL_ONE, Gl.GL_ZERO );

			Gl.glPopMatrix();

			// Draw the jumpers
			foreach ( Jumper jumper in m_jumpers )
			{
				jumper.Draw( m_light_pos );
			}

			// Draw the trampoline
			m_trampoline.Draw();

			// Draw the building
			Gl.glBegin( Gl.GL_TRIANGLE_STRIP );
			Gl.glColor3f( 0.3f, 0.0f, 0.0f );
			Gl.glVertex2f( 0, 60 );

			Gl.glColor3f( 0.1f, 0.0f, 0.0f );
			Gl.glVertex2f( 0, 480 );

			Gl.glVertex2f( 80, 60 );

			Gl.glVertex2f( 80, 480 );
			Gl.glEnd();

			// Draw the ground
			Gl.glBegin( Gl.GL_TRIANGLE_STRIP );
			Gl.glColor3f( 0.4f, 0.4f, 0.4f );
			Gl.glVertex2f( 0, 460 );

			Gl.glColor3f( 0.0f, 0.0f, 0.0f );
			Gl.glVertex2f( 0, 480 );

			Gl.glColor3f( 0.4f, 0.4f, 0.4f );
			Gl.glVertex2f( 640, 460 );

			Gl.glColor3f( 0.0f, 0.0f, 0.0f );
			Gl.glVertex2f( 640, 480 );
			Gl.glEnd();

			// Draw the particles
			m_particles.Draw();

			// Draw the HUD
			Gl.glBlendFunc( Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA );

			Gl.glColor4f( 1.0f, 1.0f, 1.0f, 0.7f );
			Fonter.Write( 500, 30, 18, "Score: " + m_score );
			Fonter.Write( 500, 42, 18, "High : " + m_high );
			Fonter.Write( 500, 58, 18, "Lives: " + m_lives );

			Gl.glColor4f( 1.0f, 1.0f, 1.0f, 0.3f );
			Fonter.Write( 500, 75, 10, "Bounced: " + m_bounces );
			Fonter.Write( 500, 83, 10, "Saved  : " + m_saved );

			if ( m_lives <= 0 )
			{
				Gl.glColor4f( 0.8f, 0.2f, 0.2f, 0.7f );
				Fonter.Write( 60, 240, 100, "Game over!" );

				Gl.glColor4f( 1.0f, 1.0f, 1.0f, 0.7f );
				Fonter.Write( 90, 430, 40.0f, "Click mouse to continue" );
			}

			Gl.glBlendFunc( Gl.GL_ONE, Gl.GL_ZERO );
		}

		public void KeyDown(SdlDotNet.Key key)
		{
			if ( key == Key.LeftArrow || key == Key.A )
			{
				m_trampoline.pos_idx = Math.Max( m_trampoline.pos_idx - 1, 0 );
			}

			if ( key == Key.RightArrow || key == Key.D )
			{
				m_trampoline.pos_idx = Math.Min( m_trampoline.pos_idx + 1, 2 );
			}
		}

		public void KeyUp(SdlDotNet.Key key)
		{
		}
	}

	public enum JumperType
	{
		Black,
		White,
	}

	public enum JumperStatus
	{
		Normal,
		OffScreen,
		Bounced,
		Killed,
	}

	public class Jumper
	{
		public JumperType type;
		public Vector2 pos;
		public Vector2 vel;

		public float ang;
		public float ang_speed;

		private static bool m_last_black = false;

		public Jumper()
		{
//			type = ( Resources.random.Next( 2 ) == 0 ) ? JumperType.White : JumperType.Black;
			m_last_black = !m_last_black;
			type = ( m_last_black ) ? JumperType.Black : JumperType.White;

			pos = new Vector2( 40.0f, 60.0f + (float)Resources.random.NextDouble() * 200 );
			vel = new Vector2( 0.8f + (float)Resources.random.NextDouble() * 1.2f, 0.0f );

			ang_speed = 2.0f + (float)Resources.random.NextDouble() * 6.0f;
			if ( Resources.random.Next( 2 ) == 1 )
				ang_speed = -ang_speed;
		}

		public JumperStatus Update( Trampoline trampoline )
		{
			vel.Y += 0.05f;
			pos += vel;
			ang += ang_speed;

			// Bounce off the trampoline?
			if ( vel.Y > 0 && trampoline.Bounce( pos ) )
			{
				vel.X *= 1.1f;
				vel.Y = -vel.Y;

				ang_speed = 2.0f + (float)Resources.random.NextDouble() * 6.0f;
				if ( Resources.random.Next( 2 ) == 1 )
					ang_speed = -ang_speed;
				return JumperStatus.Bounced;
			}

			// Squish to the ground?
			if ( pos.Y > 460 && vel.Y > 0 )
			{
				return JumperStatus.Killed;
			}

			return ( pos.X < 640 ) ? JumperStatus.Normal : JumperStatus.OffScreen;
		}

		public void Draw( Vector2 light_pos )
		{
			// Calculate how visible the character will be
			float dist_to_light = Vector2.Distance( light_pos, pos );

			float vis_mul = dist_to_light / ( GameConstants.LightRadius );
			if ( vis_mul > 1.0f )
				vis_mul = 1.0f;
			vis_mul = 1.0f - vis_mul;

			if ( vis_mul > 0.2f )
				vis_mul = 1.0f;
			else
				vis_mul /= 0.2f;

			if ( type == JumperType.White )
				vis_mul = 1.0f - vis_mul;

			// Visible near the left border (always)
			if ( pos.X < 250 )
			{
				float vis2 = 1.0f - (pos.X / 250);
				if ( vis2 < 0.0f )
					vis2 = 0.0f;
				if ( vis2 > 1.0f )
					vis2 = 1.0f;

				vis_mul = vis2 * 1.0f + (1.0f - vis2 ) * vis_mul;
			}

			// Visible?
			float min_vis = ( type == JumperType.Black ) ? 0.05f : 0.05f;

			if ( vis_mul < min_vis )
			{
				vis_mul = min_vis;
				return;
			}

			float col = ( type == JumperType.Black ) ? 0.2f : 0.6f;

			// Draw it
			Gl.glPushMatrix();
			Gl.glTranslatef( pos.X, pos.Y, 0 );
			Gl.glRotatef( ang, 0, 0, 1 );

			Gl.glBlendFunc( Gl.GL_SRC_ALPHA, Gl.GL_ONE_MINUS_SRC_ALPHA );
			Gl.glLineWidth( 2.0f );

			Gl.glBegin( Gl.GL_LINES );
			Gl.glColor4f( col, col, col, vis_mul );

			// (body)
			Gl.glVertex2f( 0, -5 );
			Gl.glVertex2f( 0, 5 );

			// (leg)
			Gl.glVertex2f( 0, 5 );
			Gl.glVertex2f( -6, 10 );

			// (leg)
			Gl.glVertex2f( 0, 5 );
			Gl.glVertex2f( 6, 10 );

			// (arm)
			Gl.glVertex2f( 0, 0 );
			Gl.glVertex2f( -6, -4 );

			// (arm)
			Gl.glVertex2f( 0, 0 );
			Gl.glVertex2f( 6, -4 );

			// (head)
			Gl.glVertex2f( -2, -5 );
			Gl.glVertex2f( 2, -5 );

			Gl.glEnd();

			Gl.glBlendFunc( Gl.GL_ONE, Gl.GL_ZERO );

			Gl.glPopMatrix();
		}
	}
}
